#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "chatform.h"
#include "xmppthread.h"
#include "rosteritem.h"
#include <QStandardItemModel>
#include <QMessageBox>
#include <QToolButton>
#include <QComboBox>

MainWindow::MainWindow(QWidget *parent) :
		QMainWindow(parent),
		ui(new Ui::MainWindow),
		mThread(NULL),
		mModel(NULL)
{
	ui->setupUi(this);

	connect(ui->lineEdit_Username, SIGNAL(returnPressed()), this, SLOT(on_pushButton_clicked()));
	connect(ui->lineEdit_Password, SIGNAL(returnPressed()), this, SLOT(on_pushButton_clicked()));
	connect(ui->actionQuit, SIGNAL(triggered()), qApp, SLOT(quit()));

	mDomains.push_back("gmail.com");
	mDomains.push_back("jabber.org");

	mServers.push_back("talk.google.com");
	mServers.push_back("jabber.org");
}

MainWindow::~MainWindow()
{
	delete ui;

	if (mThread != NULL)
	{
		disconnect(mThread, NULL, NULL, NULL);
		mThread->quit();
		mThread->wait();

		delete mThread;
		mThread = NULL;
	}
}

void MainWindow::changeEvent(QEvent *e)
{
	QMainWindow::changeEvent(e);

	switch (e->type())
	{
	case QEvent::LanguageChange:
		ui->retranslateUi(this);
		break;
	default:
		break;
	}
}

void MainWindow::closeEvent(QCloseEvent *e)
{
	closeAllChatForms();
	QMainWindow::closeEvent(e);
}

void MainWindow::on_pushButton_clicked()
{
	ui->lineEdit_Username->setEnabled(false);
	ui->lineEdit_Password->setEnabled(false);
	ui->comboBox_Server->setEnabled(false);
	ui->pushButton->setEnabled(false);
	ui->comboBox_Status->setCurrentIndex(0);

	mThread = new XmppThread(this,
							ui->lineEdit_Username->text(),
							mDomains[ui->comboBox_Server->currentIndex()],
							ui->lineEdit_Password->text(),
							mServers[ui->comboBox_Server->currentIndex()]);

	connect(mThread, SIGNAL(finished()), this, SLOT(on_xmppThread_finished()));
	connect(mThread, SIGNAL(authSuccessful()), this, SLOT(onAuthSuccessful()));
	connect(mThread, SIGNAL(authFailed()), this, SLOT(onAuthFailed()));
	connect(mThread, SIGNAL(rosterItemReceived(QString,QString)), this, SLOT(onRosterItemReceived(QString,QString)));
	connect(mThread, SIGNAL(statusReceived(QString,QString,QString)), this, SLOT(onStatusReceived(QString,QString,QString)));
	connect(mThread, SIGNAL(messageReceived(QString,QString)), this, SLOT(onMessageReceived(QString,QString)));

	mThread->start(QThread::IdlePriority);
}

void MainWindow::on_xmppThread_finished()
{
	delete mThread;
	mThread = NULL;

	closeAllChatForms();
	ui->listView->setModel(NULL);

	delete mModel;
	mModel = NULL;

	ui->lineEdit_Username->setEnabled(true);
	ui->lineEdit_Password->setEnabled(true);
	ui->comboBox_Server->setEnabled(true);
	ui->pushButton->setEnabled(true);
	ui->stackedWidget->setCurrentIndex(0);
}

void MainWindow::on_listView_doubleClicked(const QModelIndex &index)
{
	RosterItem *item = static_cast<RosterItem *>(mModel->item(index.row()));
	showChatForm(item);
}

void MainWindow::showChatForm(RosterItem *item)
{
	Q_ASSERT(item != NULL);

	ChatForm *form = item->form();

	if (form == NULL)
	{
		form = new ChatForm();
		form->setJid(item->jid());
		form->setWindowTitle(item->text());
		item->setForm(form);
		connect(form, SIGNAL(messageEntered(QString,QString)), mThread, SLOT(sendMessage(QString,QString)));
	}

	form->show();
	form->setFocus();
}

void MainWindow::closeAllChatForms()
{
	if (mModel != NULL)
	{
		for (int row = 0; row < mModel->rowCount(); ++row)
		{
			RosterItem *item = static_cast<RosterItem *>(mModel->item(row));
			ChatForm *form = item->form();

			if (form != NULL)
				form->close();
		}
	}
}

RosterItem* MainWindow::getRosterItemByJid(const QString &jid)
{
	for (int row = 0; row < mModel->rowCount(); ++row)
	{
		RosterItem *item = static_cast<RosterItem *>(mModel->item(row));

		if (item->jid() == jid)
			return item;
	}

	return NULL;
}

void MainWindow::on_comboBox_Status_currentIndexChanged(int index)
{
	QString status;

	switch (index)
	{
	case 0:
		status = "available";
		break;
	case 1:
		status = "away";
		break;
	case 2:
		status = "dnd";
		break;
	case 3:
		status = "unavailable";
		break;
	}

	if (mThread != NULL)
		mThread->sendStatus(status, "");
}

void MainWindow::onAuthSuccessful()
{
	mModel = new QStandardItemModel(this);
	ui->listView->setModel(mModel);
	ui->stackedWidget->setCurrentIndex(1);
}

void MainWindow::onAuthFailed()
{
	QMessageBox::warning(this, tr("QTalk"), tr("Invalid username or password."));

	ui->lineEdit_Username->setEnabled(true);
	ui->lineEdit_Password->setEnabled(true);
	ui->comboBox_Server->setEnabled(true);
	ui->pushButton->setEnabled(true);
}

void MainWindow::onRosterItemReceived(QString jid, QString name)
{
	RosterItem *item = new RosterItem(jid.toLower(), name);
	mModel->appendRow(item);
	mModel->sort(0);
}

void MainWindow::onStatusReceived(QString from, QString status, QString message)
{
	QStringList list = from.split('/');
	RosterItem *item = getRosterItemByJid(list[0].toLower());

	if (item != NULL)
	{
		item->setStatus(status, message);
		mModel->sort(0);
	}
}

void MainWindow::onMessageReceived(QString from, QString message)
{
	QStringList list = from.split('/');
	RosterItem *item = getRosterItemByJid(list[0].toLower());

	if (item != NULL)
	{
		showChatForm(item);
		item->form()->addMessage(message);
		QApplication::alert(item->form());
	}
}

void MainWindow::on_actionAbout_triggered()
{
	QString text = tr(
			"A very simple Jabber client that connects to your Google Talk account.<br><br>"
			"Implemented by Jakov Fustin. Visit <a href='http://fust.in'>fust.in</a> to see more of my work.");

	QMessageBox::about(this, tr("About QTalk"), text);
}

void MainWindow::on_actionAboutQt_triggered()
{
	QMessageBox::aboutQt(this, tr("About Qt"));
}
