#include "xmppclient.h"
#include "xmlhandler.h"
#include "xmlelement.h"
#include <QCryptographicHash>

XmppClient::XmppClient() :
	mXmlReader(NULL),
	mXmlHandler(NULL),
	mNewStreamInitiated(true),
	mUserAuthenticated(false),
	mSaslResponseSent(false),
	mResourceBound(false),
	mSessionEstablished(false),
	mBindId("bind_1"),
	mSessionId("session_1"),
	mRosterId("roster_1")
{
	mSocket = new QSslSocket(this);
	mXmlSource = new QXmlInputSource;

	connect(mSocket, SIGNAL(connected()), this, SLOT(initiateNewStream()));
	connect(mSocket, SIGNAL(disconnected()), this, SIGNAL(disconnected()));
	connect(mSocket, SIGNAL(encrypted()), this, SLOT(initiateNewStream()));
	connect(mSocket, SIGNAL(readyRead()), this, SLOT(readFromSocket()));
}

XmppClient::~XmppClient()
{
	delete mXmlHandler;
	delete mXmlReader;
	delete mXmlSource;
	delete mSocket;

	mXmlHandler = NULL;
	mXmlReader = NULL;
	mXmlSource = NULL;
	mSocket = NULL;
}

void XmppClient::connectToHost(const QString &username,
							   const QString &domain,
							   const QString &password,
							   const QString &hostname,
							   ushort port)
{
	mUsername = username;
	mDomain = domain;
	mPassword = password;

	mSocket->connectToHost(hostname, port);
}

void XmppClient::disconnectFromHost()
{
	mSocket->disconnectFromHost();
}

void XmppClient::sendMessage(QString to, QString message)
{
	QString msg = "<message to='%1' from='%2' type='chat' xml:lang='en'>"
				  "<body>%3</body></message>";

	sendMessageToHost(msg.arg(to, mJid, message));
}

void XmppClient::sendStatus(QString status, QString message)
{
	XmlElement *presence = new XmlElement("presence");

	if(status == "unavailable")
	{
		presence->setAttribute("type", status);
	}
	else if(status == "away" || status == "xa" || status == "dnd")
	{
		XmlElement *show = new XmlElement("show");
		show->setData(status);
		presence->addChild(show);
	}

	if(message.length() > 0)
	{
		XmlElement *statusMsg = new XmlElement("status");
		statusMsg->setData(message);
		presence->addChild(statusMsg);
	}

	sendMessageToHost(presence->toString());

	if(status == "unavailable")
		sendMessageToHost("</stream:stream>");

	delete presence;
}

void XmppClient::sendMessageToHost(const QString &msg)
{
	mSocket->write(msg.toUtf8());
}

void XmppClient::initiateNewStream()
{
	delete mXmlHandler;
	delete mXmlReader;

	mXmlHandler = new XmlHandler;
	mXmlReader = new QXmlSimpleReader;
	mXmlReader->setContentHandler(mXmlHandler);

	connect(mXmlHandler, SIGNAL(streamStarted(const XmlElement*)), this, SLOT(onStreamStarted(const XmlElement*)));
	connect(mXmlHandler, SIGNAL(streamEnded()), this, SLOT(onStreamEnded()));
	connect(mXmlHandler, SIGNAL(elementReady(const XmlElement*)), this, SLOT(queueElement(const XmlElement*)));

	mNewStreamInitiated = true;

	QString msg = "<?xml version='1.0'?><stream:stream xmlns='jabber:client' "
				  "xmlns:stream='http://etherx.jabber.org/streams' "
				  "to='%1' version='1.0'>";

	sendMessageToHost(msg.arg(mDomain));
}

void XmppClient::readFromSocket()
{
	bool ok;

	mXmlSource->setData(mSocket->readAll());

	if(mNewStreamInitiated)
	{
		mNewStreamInitiated = false;
		ok = mXmlReader->parse(mXmlSource, true);
	}
	else
	{
		ok = mXmlReader->parseContinue();
	}

	if(!ok)
		disconnectFromHost();

	while(mQueue.size() > 0)
		processQueuedData();
}

void XmppClient::onStreamStarted(const XmlElement *elem)
{
	Q_ASSERT(elem != NULL);
}

void XmppClient::onStreamEnded()
{
	// Insert special element that denotes the end of stream
	queueElement(new XmlElement("END_OF_STREAM"));
}

void XmppClient::queueElement(const XmlElement *elem)
{
	Q_ASSERT(elem != NULL);
	mQueue.insert(mQueue.begin(), new XmlElement(*elem));
}

void XmppClient::processQueuedData()
{
	XmlElement* elem = mQueue.back();

	if(elem != NULL)
	{
		QString tagName = elem->tagName();

		if(tagName == "END_OF_STREAM")
		{
			disconnectFromHost();
		}
		else if(tagName == "http://etherx.jabber.org/streams:features")
		{
			processStreamFeatures(elem);
			selectStreamFeatures();
		}
		else if(tagName == "urn:ietf:params:xml:ns:xmpp-tls:proceed")
		{
			mSocket->setProtocol(QSsl::TlsV1);
			mSocket->setPeerVerifyMode(QSslSocket::VerifyNone);
			mSocket->startClientEncryption();
		}
		else if(tagName == "urn:ietf:params:xml:ns:xmpp-tls:failure")
		{
		}
		else if(tagName == "urn:ietf:params:xml:ns:xmpp-sasl:challenge")
		{
			QByteArray challenge = QByteArray::fromBase64(elem->data().toUtf8());
			processSaslChallenge(QString(challenge));
		}
		else if(tagName == "urn:ietf:params:xml:ns:xmpp-sasl:success")
		{
			mUserAuthenticated = true;
			emit authSuccessful();
			initiateNewStream();
		}
		else if(tagName == "urn:ietf:params:xml:ns:xmpp-sasl:failure")
		{
			emit authFailed();
		}
		else if(tagName == "jabber:client:iq")
		{
			processIqStanza(elem);
		}
		else if(tagName == "jabber:client:message")
		{
			processMessageStanza(elem);
		}
		else if(tagName == "jabber:client:presence")
		{
			processPresenceStanza(elem);
		}
	}

	// Remove processed data from queue
	delete elem;
	mQueue.pop_back();
}

void XmppClient::selectStreamFeatures()
{
	if(!mSocket->isEncrypted() && mStreamFeatures.TlsAvailable)
	{
		sendMessageToHost("<starttls xmlns='urn:ietf:params:xml:ns:xmpp-tls' />");
	}
	else if(!mUserAuthenticated)
	{
		if(mStreamFeatures.SaslDigestMD5Available)
		{
			sendMessageToHost("<auth xmlns='urn:ietf:params:xml:ns:xmpp-sasl' "
							  "mechanism='DIGEST-MD5' />");
		}
		else if(mStreamFeatures.SaslPlainAvailable)
		{
			QString msg = "<auth xmlns='urn:ietf:params:xml:ns:xmpp-sasl' "
						  "mechanism='PLAIN'>%1</auth>";

			QString auth = '\0' + mUsername + '\0' + mPassword;

			sendMessageToHost(msg.arg(QString(auth.toUtf8().toBase64())));
		}
		else
		{
			// Error! Other SASL mechanisms aren't supported.
		}
	}
	else if(!mResourceBound && mStreamFeatures.BindingRequired)
	{
		QString msg = "<iq type='set' id='%1'>"
					  "<bind xmlns='urn:ietf:params:xml:ns:xmpp-bind' /></iq>";

		sendMessageToHost(msg.arg(mBindId));
	}
	else if(!mSessionEstablished && mStreamFeatures.SessionRequired)
	{
		QString msg = "<iq to='%1' type='set' id='%2'>"
					  "<session xmlns='urn:ietf:params:xml:ns:xmpp-session' /></iq>";

		sendMessageToHost(msg.arg(mDomain, mSessionId));
	}
}

void XmppClient::processStreamFeatures(const XmlElement *features)
{
	Q_ASSERT(features != NULL);

	XmlElement *elem;

	elem = features->child("urn:ietf:params:xml:ns:xmpp-tls:starttls");
	if(elem != NULL)
	{
		mStreamFeatures.TlsAvailable = true;
	}

	elem = features->child("urn:ietf:params:xml:ns:xmpp-sasl:mechanisms");
	if(elem != NULL)
	{
		for (uint i=0; i<elem->numOfChildren(); ++i)
		{
			if(elem->child(i)->tagName() == "urn:ietf:params:xml:ns:xmpp-sasl:mechanism")
			{
				QString data = elem->child(i)->data();

				if(data == "PLAIN")
					mStreamFeatures.SaslPlainAvailable = true;
				else if(data == "DIGEST-MD5")
					mStreamFeatures.SaslDigestMD5Available = true;
			}
		}
	}

	elem = features->child("urn:ietf:params:xml:ns:xmpp-bind:bind");
	if(elem != NULL)
	{
		if(elem->data() == "")
			mStreamFeatures.BindingRequired = true;
	}

	elem = features->child("urn:ietf:params:xml:ns:xmpp-session:session");
	if(elem != NULL)
	{
		if(elem->data() == "")
			mStreamFeatures.SessionRequired = true;
	}
}

void XmppClient::processSaslChallenge(const QString &challenge)
{
	if (!mSaslResponseSent)
	{
		// We use this digest to authenticate ourselves
		QString response = digestMd5Response(challenge, true);

		// This digest is used to make sure the server isn't messing with us
		mExpectedAuthResponse = digestMd5Response(challenge, false);

		QString msg = "<response xmlns='urn:ietf:params:xml:ns:xmpp-sasl'>%1</response>";

		mSaslResponseSent = true;
		sendMessageToHost(msg.arg(QString(response.toUtf8().toBase64())));
	}
	else
	{
		QStringList list = challenge.split('=');

		if (mExpectedAuthResponse != list[1])
		{
			// Oops! Server miscalculated the hash. Perhaps somebody's
			// messing with us. It's probably best to disconnect now.
			disconnectFromHost();
		}

		QString msg = "<response xmlns='urn:ietf:params:xml:ns:xmpp-sasl' />";
		sendMessageToHost(msg);
	}
}

QString XmppClient::digestMd5Response(const QString &challenge, bool auth)
{
	QString realm, nonce;
	QString cnonce = "U8L0SDh6VqTrRk";

	QStringList keyValuePairs = challenge.split(',');

	for (int i = 0; i < keyValuePairs.length(); ++i)
	{
		QStringList pair = keyValuePairs[i].split('\"');

		if (pair[0] == "realm=")
			realm = pair[1];
		else if (pair[0] == "nonce=")
			nonce = pair[1];
	}

	// Build a digest
	QString a1(mUsername + ":" + mDomain + ":" + mPassword);
	QString a2(":xmpp/" + mDomain);

	if (auth)
		a2 = "AUTHENTICATE" + a2;

	QCryptographicHash::Algorithm md5 = QCryptographicHash::Md5;

	QString Ha1(QCryptographicHash::hash(a1.toAscii(), md5));
	Ha1 += ":" + nonce + ":" + cnonce;

	QString hexHa1(QCryptographicHash::hash(Ha1.toAscii(), md5).toHex());
	QString hexHa2(QCryptographicHash::hash(a2.toAscii(), md5).toHex());
	QString middle = ":" + nonce + ":00000001:" + cnonce + ":auth:";
	QString total = hexHa1 + middle + hexHa2;
	QString digest(QCryptographicHash::hash(total.toAscii(), md5).toHex());

	// Build a response
	QString response;

	if (auth)
	{
		response = "username=\"" + mUsername +
				   "\",realm=\"" + mDomain +
				   "\",nonce=\"" + nonce +
				   "\",cnonce=\"" + cnonce +
				   "\",nc=00000001,qop=auth,digest-uri=\"xmpp/" + mDomain +
				   "\",response=" + digest + ",charset=utf-8";
	}
	else
	{
		response = digest;
	}

	return response;
}

void XmppClient::processIqStanza(const XmlElement *iq)
{
	Q_ASSERT(iq != NULL);

	if(iq->hasAttribute("id") && iq->hasAttribute("type"))
	{
		QString id = iq->attribute("id");
		QString type = iq->attribute("type");

		if(type == "result")
		{
			if(id == mBindId)
			{
				XmlElement *jid = iq->child("urn:ietf:params:xml:ns:xmpp-bind:jid");

				if(jid != NULL)
				{
					mJid = jid->data();
					mResourceBound = true;

					selectStreamFeatures();
				}
			}
			else if(id == mSessionId)
			{
				mSessionEstablished = true;

				QString msg = "<iq from='%1' type='get' id='%2'>"
							  "<query xmlns='jabber:iq:roster' /></iq>";

				sendMessageToHost(msg.arg(mJid, mRosterId));
			}
			else if(id == mRosterId)
			{
				processRoster(iq);
				sendMessageToHost("<presence />");
			}
		}
	}
}

void XmppClient::processMessageStanza(const XmlElement *msg)
{
	Q_ASSERT(msg != NULL);

	if(msg->hasAttribute("from"))
	{
		QString from;
		QString body;

		from = msg->attribute("from");

		XmlElement* elem = msg->child("jabber:client:body");

		if(elem != NULL)
			body = elem->data();

		emit messageReceived(from, body);
	}
}

void XmppClient::processPresenceStanza(const XmlElement *presence)
{
	Q_ASSERT(presence != NULL);

	if(presence->hasAttribute("from"))
	{
		QString from = presence->attribute("from");

		QString type;
		if(presence->hasAttribute("type"))
			type = presence->attribute("type");

		XmlElement* elem;

		QString status;
		elem = presence->child("jabber:client:show");
		if(elem != NULL)
		{
			status = elem->data();
		}

		QString message;
		elem = presence->child("jabber:client:status");
		if(elem != NULL)
		{
			message = elem->data();
		}

		if(type == "" || type == "unavailable")
		{
			if(type == "" && status == "")
			{
				status = "available";
			}
			else if(type == "unavailable")
			{
				status = "unavailable";
			}

			emit statusReceived(from, status, message);
		}
		else
		{
			// Other presence stanzas (subscribe, unsubscribe etc.) need
			// to be handled here.
		}
	}
}

void XmppClient::processRoster(const XmlElement *iq)
{
	Q_ASSERT(iq != NULL);

	XmlElement* elem = iq->child("jabber:iq:roster:query");

	if(elem != NULL)
	{
		for (uint i=0; i<elem->numOfChildren(); ++i)
		{
			if(elem->child(i)->tagName() == "jabber:iq:roster:item")
			{
				QString jid;
				QString name;
				QString subscription;

				if(elem->child(i)->hasAttribute("jid"))
				{
					jid = elem->child(i)->attribute("jid");

					if(elem->child(i)->hasAttribute("name"))
						name = elem->child(i)->attribute("name");

					if(elem->child(i)->hasAttribute("subscription"))
						subscription = elem->child(i)->attribute("subscription");

					emit rosterItemReceived(jid, name);
				}
			}
		}
	}
}
